/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Copyright (C) 2002  Martijn Leisink  [martijn@mbfys.kun.nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "boundprop.h"
#include <iostream>


namespace libDAI {


    using namespace std;


    // Set up LP problem that restricts marginals on sep based on this set of bounds
    void BoundSet::make_lp( const VarSet& sep ) {
        size_t n = sep.states();    // number of columns

        size_t m = 0;
        for( const_iterator b = begin(); b != end(); ++b )
            if( b->vars() << sep ) 
                m += b->vars().states();
        if( Verbose() >= 2 )
            cerr << "Setting up LP problem for " << sep << " with " << m << " inequality constraints and " << n << " variables" << endl;
        lpx = ::lpx_create_prob();
        ::lpx_add_rows( lpx, m+1 );
        ::lpx_add_cols( lpx, n );

        int *ind = new int[n+1];
        double *val = new double[n+1];
        for( size_t j = 1; j <= n; j++ ) {
            ind[j] = j;  val[j] = 1.0;
        }
        ::lpx_set_mat_row( lpx, m+1, n, ind, val );
        delete val;
        delete ind;
        ::lpx_set_row_bnds( lpx, m+1, LPX_FX, 1.0, 1.0 );

        m = 0;
        for( const_iterator b = begin(); b != end(); b++ )
            if( b->vars() << sep ) {
                IndexFor i( b->vars(), b->vars() );
                IndexFor j( sep, b->vars() );
                for( ; i >= 0; ++i, ++j, m++ ) {
                    size_t len = (sep - b->vars()).states();
                    int *ind = new int[len+1];
                    double *val = new double[len+1];
                    IndexFor k( sep, sep - b->vars() );
                    for( size_t l = 1; k >= 0; ++k, l++ ) {
                        ind[l] = 1 + j + k;
                        val[l] = 1.0;
                        /* a_{m+1,ind[l]} = val[l] */
                    }
                    ::lpx_set_mat_row( lpx, m+1, len, ind, val );
                    delete val;
                    delete ind;
                    ::lpx_set_row_bnds( lpx, m+1, LPX_DB, b->min()[i], b->max()[i] );
                }
            }

        for( size_t j = 0; j < n; j++ )
            ::lpx_set_col_bnds( lpx, j+1, LPX_LO, 0.0, 0.0 );

        lpx_set_int_parm( lpx, LPX_K_MSGLEV, 0 );      // quiet
    }

    // Improve bound b
    void BoundSet::newbound (Box& b, const Factor& Pjoint, const Factor& Psep) const {
        VarSet other = Pjoint.vars() - (Psep.vars() + b.vars());
        IndexFor i_joint(Pjoint.vars(),b.vars());
        IndexFor i_bound(b.vars(),b.vars());
        for(;i_joint>=0;++i_joint,++i_bound) {
            IndexFor j_joint(Pjoint.vars(),Psep.vars());
            IndexFor j_sep(Psep.vars(),Psep.vars());
            int res = 0;
            double result;

            for(;j_sep>=0;++j_sep,++j_joint) {
                double x=0;
                IndexFor k_joint(Pjoint.vars(),other);
                for(;k_joint>=0;++k_joint) 
                    x += Pjoint[i_joint + j_joint + k_joint];
                double val = 0.0;
                if( Psep[j_sep] != 0.0 )
                    val = x / Psep[j_sep];
                ::lpx_set_obj_coef( lpx, 1 + j_sep, val );
            }

            lpx_set_obj_dir( lpx, LPX_MAX );
            res = lpx_simplex( lpx );
            if( lpx_get_status( lpx ) != LPX_OPT ) {
                cerr << "WARNING: lpx_simplex did not find an optimum (return value " << res << "), ignoring this result" << endl;
            } else {
                result = lpx_get_obj_val( lpx );
                if( result < b.max()[i_bound] )
                    b.max()[i_bound] = result;
            }

            lpx_set_obj_dir( lpx, LPX_MIN );
            res = lpx_simplex( lpx );
            if( lpx_get_status( lpx ) != LPX_OPT ) {
                cerr << "WARNING: lpx_simplex did not find an optimum (return value " << res << "), ignoring this result" << endl;
            } else {
                result = lpx_get_obj_val( lpx );
                if( result > b.min()[i_bound] )
                    b.min()[i_bound] = result;
            }
        }
    }


    BoundSet BoundSet::up (const FactorGraph& net) {
        BoundSet nbs;
//        cerr<<"Improving bound set "; cerr.flush();
        for(const_iterator b=begin();b!=end();++b) {
            VarSet sep = net.delta( b->vars() );

            //Factor Pjoint = net.markov_blanket((NodeSet)*b);
            Factor Pjoint;
            for( size_t I = 0; I < net.nrFactors(); I++ )
                if( net.factor(I).vars().intersects( b->vars() ) )
                    Pjoint *= net.factor(I);

            Box nb(b->vars());

            make_lp(sep);
            newbound(nb,Pjoint,Pjoint.partSum(sep));
            delete_lp();

            nbs.insert( nb );
        }
//        cerr<<endl;
        return(nbs);
    }


    BoundSet BoundSet::up (const FactorGraph& net, const Cluster& base) {
        BoundSet nbs(*this);
//        cerr<<"Improving bound set "; cerr.flush();
        for(Cluster::const_iterator b_base=base.begin();b_base!=base.end();++b_base) {
            VarSet sep = net.delta(b_base->vars());
            make_lp(sep);
            //Factor Pjoint = net.markov_blanket((NodeSet)*b_base);
            Factor Pjoint;
            for( size_t I = 0; I < net.nrFactors(); I++ )
                if( net.factor(I).vars().intersects( b_base->vars() ) )
                    Pjoint *= net.factor(I);
            Factor Psep = Pjoint.partSum(sep);
            for(const_iterator b=nbs.begin();b!=nbs.end();)
                if( b->vars() << b_base->vars() )
                {
                    Box nb(b->vars());
                    ++b; // iterator is invalid after nbs.erase(nb)
                    nbs.erase(nb);
                    newbound(nb,Pjoint,Psep);
                    nbs.insert(nb);
                } else ++b;
            delete_lp();
        }
//        cerr<<endl;
        return(nbs);
    }


    void BoundProp_solve( const FactorGraph &fg, long maxsize, double tol, size_t maxiter, BoundSet &bounds, size_t verb ) {
        Cluster Bmain;
        BoundProp_setup( fg, maxsize, bounds, Bmain, verb );

        double max = 1.0;
        size_t iter = 1;
        while( max > tol && iter < maxiter ) {
            if( bounds.Verbose() >= 2 ) {
                cerr << "==> Using " << bounds.size() << " bounding potentials" << endl;
                cerr << "==> Iteration " << iter << endl;
            }
            BoundSet bounds_new = bounds.up( fg, Bmain );
            if( bounds.Verbose() >= 2 ) {
                cerr << "==> Calculating delta" << endl;
            }

            max = 0.0;
            BoundSet::iterator it_new = bounds_new.begin();
            for( BoundSet::iterator it = bounds.begin(); it != bounds.end(); it++, it_new++ ) {
                double delta_old = it->gap();
                double delta_new = it_new->gap();
                if( delta_old - delta_new > max )
                    max = delta_old - delta_new;
            }
            if( bounds.Verbose() >= 1 ) {
                cerr.precision( 10 );
                cerr << "# " << iter << " " << max << endl;
            }
            bounds = bounds_new;
            iter++;
        }
    }


    void BoundProp_setup( const FactorGraph &fg, long maxsize, BoundSet &B, Cluster &Bmain, size_t verb ) {
        if( maxsize > 0 ) {
            Cluster Csub;
            BoundProp_tile( fg, maxsize, Bmain, Csub );
    //        Cluster all = Csub;
    //        set_union( Bmain->begin(), Bmain->end(), Csub.begin(), Csub.end(), inserter( all, all.begin() ) );
            for( Cluster::const_iterator c = Bmain.begin(); c != Bmain.end(); c++ )
                Csub.insert( *c );
            B = BoundSet(Csub);
        } else {
            Bmain.clear();
            for( VarSet::const_iterator n = fg.vars().begin(); n != fg.vars().end(); n++ )
                Bmain.insert( Factor( *n ) );
            B = BoundSet(Bmain);
        }
        B.Verbose( verb );
    }


    void tiling( Cluster& C, const FactorGraph & N, const VarSet& ns, const VarSet& sep, long maxsize, size_t Verbose ) {
        if( maxsize == 0 || (maxsize > 0 && sep.states() * ns.states() > (unsigned long)maxsize) )
            return;
        if( Verbose ) {
            cerr << "."; 
            cerr.flush();
        }
        if( !C.insert(ns).second ) 
            return;
        if( maxsize < 0 ) 
            maxsize++;
        for( VarSet::const_iterator n = sep.begin(); n != sep.end(); ++n )
            if( !ns.contains( *n ) ) {
                VarSet newns = ns + *n;
                VarSet newsep = (sep + N.delta(*n)) - newns;
                tiling( C, N, newns, newsep, maxsize, Verbose );
            }
    }


    void split( Cluster& Cinside, Cluster& C ) {
        for( Cluster::iterator i = C.begin(); i != C.end(); ) {
            Cluster::iterator j;
            for( j = C.begin(); j != C.end() && !((i->vars() << j->vars()) && i->vars() != j->vars()); ++j );
            if( j != C.end() ) { 
                Cinside.insert(*i); 
                j = i; 
                ++i; 
                C.erase( *j ); 
            } else 
                ++i;
        }
    }


    void BoundProp_tile( const FactorGraph &fg, long maxsize, Cluster &Cex, Cluster &Cin, size_t Verbose ) {
        /*
    %TILING    Find all node clusters upto a given state space size.
    %   TILING(NET,MAXSIZE) returns all sets of nodes for which the state
    %   space of the node set and its separator is less than or equal to
    %   MAXSIZE.  If MAXSIZE<0 there are at most -MAXSIZE nodes included
    %   in the node set regardless of the size of their separator.
    %
    %   [Cex,Cin] = TILING(NET,MAXSIZE) returns a set Cin of all node
    %   sets which are entirely included by at least one of the node sets
    %   in Cex.  The union of Cex and Cin is equivalent to the return
    %   value of the call in the previous paragraph.
    */

        Cluster C;
        if( Verbose ) {
            cerr << "Tiling "; 
            cerr.flush();
        }
        for( VarSet::const_iterator n = fg.vars().begin(); n != fg.vars().end(); ++n )
            tiling( C, fg, *n, fg.delta(*n), maxsize, Verbose );
        if( Verbose )
            cerr << endl;
        Cluster Cinside;
        split( Cinside, C );
        Cex = C;
        Cin = Cinside;
    }


}
