/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "box.h"
#include "factorgraph.h"
#include <vector>
#include <set>
#include <algorithm>
#ifdef GLPK
    #include <glpk.h>
#endif


namespace libDAI {

    
    using namespace std;


//#define DEBUG_AR


#ifdef GLPK
    // Test if point is in the convex hull of points by solving a linear problem
    // see http://www.cs.mcgill.ca/~fukuda/soft/polyfaq/node22.html#polytope:Vredundancy
    bool TestIfPointInConvexHull( vector<Prob> points, Prob point ) {
        using namespace std;
        using namespace libDAI;

        glp_prob *lp = glp_create_prob();
        glp_set_obj_dir(lp, GLP_MAX);

        int N = points.size();
        int d = point.size();

        int n = d + 1; // number of columns (structural variables)
        int m = N + 1; // number of rows (auxiliary variables)

        glp_add_cols(lp,n);
        glp_add_rows(lp,m);

        // set objective function
        glp_set_obj_coef(lp,n,-1.0);
        for( int j = 1; j <= d; j++ )
            glp_set_obj_coef(lp,j,point[j-1]);

        // define auxiliary variables
        // first i=1..N rows plus last row
        // last row is identical to objective function
        for( int i = 1; i <= m; i++ ) {
            vector<int> ind;
            vector<double> val;
            ind.push_back(0); val.push_back(0.0);
            for( int j = 1; j <= d; j++ ) {
                double x;
                if( i == m )
                    x = point[j-1];
                else
                    x = points[i-1][j-1];
                if( x != 0.0 ) {
                    ind.push_back(j); val.push_back(x);
                }
            }
            ind.push_back(n); val.push_back(-1.0);
            glp_set_mat_row(lp,i,ind.size()-1,&(ind[0]),&(val[0]));
        }

        // set bounds on auxiliary variables
        for( int i = 1; i <= N; i++ ) {
            glp_set_row_bnds(lp,i,GLP_UP,0.0,0.0);
        }
        glp_set_row_bnds(lp,m,GLP_UP,0.0,1.0);

        // structural variables are free
        for( int j = 1; j <= n; j++ )
            glp_set_col_bnds(lp,j,GLP_FR,0.0,0.0);

        // set parameters
    //	lpx_set_int_parm(lp, LPX_K_PRESOL, 1);      // use pre-solver
        lpx_set_int_parm(lp, LPX_K_MSGLEV, 0);      // quiet

        // solve
        if( lpx_simplex(lp) != LPX_E_OK ) {
            cerr << "WARNING: lpx_simplex did not terminate successfully!" << endl;
            assert( 0 == 1);
        }
        bool failed = (lpx_get_status(lp) != LPX_OPT);
        double result = glp_get_obj_val(lp);
        glp_delete_prob(lp);

        if( failed )
            cerr << "WARNING: solution is not optimal!" << endl;

        return (result <= 0.5);     // this may be better because of numerical roundoff errors?
//        return (result <= 0.0);
    }
#endif


    std::vector<Prob> Box::extremePoints() const {
        // FIXME: why is this disabled?
/*      bool min_is_max = false;
        if( _max.totalSum() - _min.totalSum() < 1e-14 )
            min_is_max = true;
        if( min_is_max ) {
            return std::vector<Prob>(1,_min);
        } else*/

        // THIS IS WRONG: e.g. consider [0,0,0] <= x <= [1,1,0]; the extreme points of that box have third coordinate = 0
        // FIXME: isSimplex() also doesn't work in that case...
/*      bool min_zero = false;
        if( isSimplex() )
            min_zero = true;
        if( min_zero ) { 
            size_t N = _vars.states();
            std::vector<Prob> result( N, Prob(N, 0.0) );
            for( size_t i = 0; i < N; i++ )
                result[i][i] = 1.0;
            return result;
        } else {*/
            size_t maxbits = sizeof(size_t) * 8;
            size_t N = _vars.states();
            if( N > 16 )
                std::cerr << "Warning: calculating " << (1UL << N) << " extreme points..." << std::endl;
            assert( N < maxbits );
            std::vector<Prob> result;
            result.reserve( 1UL << N );

            //for( size_t l = (min_zero ? 1 : 0); l < (1UL << N); l++ ) {
            for( size_t l = 0; l < (1UL << N); l++ ) {
                Prob point(N, 0.0);
                bool nonzero = false;
                for( size_t i = 0; i < N; i++ ) {
                    if( l & (1UL << i) )
                        point[i] = _max[i];
                    else
                        point[i] = _min[i];
                    if( point[i] )
                        nonzero = true;
                }
                if( nonzero )
                    result.push_back( point );
            }
            return result;
//        }
    }


    Box Box::normalized() const {
        // By definition, calculate bounding box of normalized projection of *this
        //
        // This calculation can actually be done coordinate-wise, which prevents
        // an expensive enumeration of extreme points
        
        if( true ) {
            // The fast way (quadratic in number of dimensions)
            Box result( _vars );
            for( size_t i = 0; i < _min.size(); i++ ) {
                // Care has to be taken to avoid numerical instability:
                // The faster alternative of calculating denom (precalculating
                // the sum over all _max[j] and then subtracting _max[i]) is
                // unfortunately numerically significantly less stable than the
                // slower method employed here
                double denom = 0.0;
                for( size_t j = 0; j < _max.size(); j++ )
                    if( j != i )
                        denom += _max[j];
                result._min[i] = _min[i] ? _min[i] / (_min[i] + denom) : 0.0;

                denom = 0.0;
                for( size_t j = 0; j < _max.size(); j++ )
                    if( j != i )
                        denom += _min[j];
                result._max[i] = _max[i] ? _max[i] / (_max[i] + denom) : 0.0;
            }
            return result;
        } else {
            // The slow way (exponential in number of dimensions)
            std::vector<Prob> ep = extremePoints();
            for( size_t l = 0; l < ep.size(); l++ )
                ep[l].normalize();
            return Box( _vars, ep );
        }
    }


    // Calculate a loose bound on the product of the boxes by
    // multiplying together their respective bounds
    Box boundProd( const std::vector<Box> &boxes ) {
        VarSet dest;
        for( size_t l = 0; l < boxes.size(); l++ )
            dest += boxes[l].vars();

        Factor min(dest, 1.0), max(dest, 1.0);
        for( size_t l = 0; l < boxes.size(); l++ ) {
            min *= Factor( boxes[l].vars(), boxes[l]._min );
            max *= Factor( boxes[l].vars(), boxes[l]._max );
        }

        return Box( dest, min.p(), max.p() );
    }

    struct less_Prob : public binary_function<size_t, size_t, bool> {
        const Prob *psi;
        less_Prob( const Prob *thePsi ) : psi(thePsi) {}
        bool operator()(size_t i, size_t j) { return psi->operator[](i) < psi->operator[](j); }
    };

    struct more_Prob : public binary_function<size_t, size_t, bool> {
        const Prob *psi;
        more_Prob( const Prob *thePsi ) : psi(thePsi) {}
        bool operator()(size_t i, size_t j) { return psi->operator[](i) > psi->operator[](j); }
    };

    // A/R algorithm by B. Tesssem ("Interval Probability Propagation" IJAR 7:95-120, 1992)
    // which is described in easier language in Draper & Hanks, "Localized Partial Evaluation of Belief Networks", UAI 1994
    //
    // Calculates a lower bound (if lower == true) or upper bound (otherwise) on (a * c).totalSum()
    // where c is the intersection of the Box b with the probability simplex.
    Real doAR( const Prob &a, const Box &b, bool lower ) {

#ifdef DEBUG_AR
        assert( a.size() == b.min().size() );
        if( !b.normalized().contains( b, 1e-9 ) )
            cerr << "b = " << b << ", b.normalized() = " << b.normalized() << endl;
        assert( b.normalized().contains( b, 1e-9 ) );
#endif

        // Create index corresponding to (ascendingly) sorted entries of a
        vector<size_t> sorted( a.size(), 0 );
        for( size_t _i = 0; _i < a.size(); _i++ )
            sorted[_i] = _i;
        if( lower )
            sort( sorted.begin(), sorted.end(), less_Prob(&a) );
        else
            sort( sorted.begin(), sorted.end(), more_Prob(&a) );

        // Initialize with lower bound of b
        Prob b_opt( b.min() );
        Real S = 1.0 - b_opt.totalSum(); // probability mass to go (Tessem calls "1-S" "S")
        for( size_t _i = 0; (_i < a.size()) && (S > 0.0); _i++ ) {
            size_t i = sorted[_i];
            b_opt[i] = std::min( b.min()[i] + S, b.max()[i] );
            S -= b.max()[i] - b.min()[i];
        }

        // The normalized element of b that optimizes the bound
        Real result = (a * b_opt).totalSum();

        // Verify if it was done correctly
#ifdef DEBUG_AR
        // For strong interactions, rounding error becomes so large that this should be disabled
        if( b.contains( b.normalized(), 1e-9 ) ) {
            if( b._min.size() <= 16 ) {
                std::vector<Prob> ep = b.extremePoints();
                Real result_min = INFINITY;
                Real result_max = -INFINITY;
                for( size_t l = 0; l < ep.size(); l++ ) {
                    Real x = (a * ep[l].normalized()).totalSum();
                    result_min = std::min( result_min, x );
                    result_max = std::max( result_max, x );
                }
                if( lower ) {
                    if( fabs( result - result_min ) >= 1e-9 )
                        cerr << "WARNING: A/R detects rounding error: " << result << " != " << result_min << endl;
                    assert( fabs( result - result_min ) < 1e-9 );
                } else {
                    if( fabs( result - result_max ) >= 1e-9 )
                        cerr << "WARNING: A/R detects rounding error: " << result << " != " << result_max << endl;
                    assert( fabs( result - result_max ) < 1e-9 );
                }
            }
        }
#endif

        return result;
    }

    // Calculate a bounding box for the (un)normalized marginal on dest of the product of elements of *this with psi
    // If variant == AR, *this is assumed to be normalized, and the AR algorithm is used (which has lower complexity) 
    //   to calculate an unnormalized marginal on dest
    // If variant == ALT, the extreme points of *this are normalized, but the marginal on dest is not normalized
    // If variant == ORG, the marginal is normalized before being bounded
    Box Box::boundSumProd( const Factor &psi, const VarSet &dest, Box::Variant variant ) const {
        if( variant == Box::AR ) {
            Prob image_min( dest.states(), INFINITY );
            Prob image_max( dest.states(), -INFINITY );
            for( State s(dest); s.valid(); s++ ) {
                Factor psi_s = psi.slice( dest, s );
                image_min[s] = doAR( psi_s.p(), *this, true );
                image_max[s] = doAR( psi_s.p(), *this, false );
            }
            return Box( dest, image_min, image_max );
        } else {
            // FIXME: do we have to store the extreme points?
            std::vector<Prob> ep = extremePoints();
            Prob image_min( dest.states(), INFINITY );
            Prob image_max( dest.states(), -INFINITY );
            for( size_t l = 0; l < ep.size(); l++ ) {
                // FIXME: this can be sped up massively by caching the index
                Prob image_ep;
                if( variant == Box::ORG )
                    image_ep = (psi * Factor(_vars,ep[l])).marginal(dest,true).p();
                else if( variant == Box::ALT )
                    image_ep = (psi * Factor(_vars,ep[l].normalized())).marginal(dest,false).p();
                image_min = libDAI::min( image_ep, image_min );
                image_max = libDAI::max( image_ep, image_max );
            }
            Box result( dest, image_min, image_max );
#ifdef DEBUG_AR
/*          if( variant == ALT ) {
                Box resultAR = boundSumProd( psi, dest, AR );
                assert( resultAR.contains( result, 1e-9 ) ); // FIXME: this doesn't seem to hold anymore...
            }
*/
#endif
            return result;
        }
    }

    // Calculates a bounding box for the marginal on dest of the product of psi with all incoming boxes.
    // If independent is true, incoming boxes are assumed to be independent.
    // All incoming boxes should have disjoint varsets.
    //
    // If variant == AR, all incoming boxes are assumed to be normalized, and the AR algorithm is used (which has lower complexity) 
    //   to calculate an unnormalized marginal on dest
    // If variant == ALT, the extreme points of the product of incoming boxes are normalized, but the marginal on dest is not normalized
    // If variant == ORG, the marginal is normalized before being bounded
    //
    // FIXME: can we improve the AR algorithm to deal with a product of independent bounded incoming measures?
    Box boundSumProd( const Factor &psi, vector<Box> &incomingBoxes, const VarSet &dest, bool independent, Box::Variant variant ) {

#ifdef LIBDAI_DEBUG
        // Check that all incomingBoxes have disjoint sets of variables
        for( size_t l = 0; l < incomingBoxes.size(); l++ )
            for( size_t k = l + 1; k < incomingBoxes.size(); k++ )
                assert( !incomingBoxes[k].vars().intersects( incomingBoxes[l].vars() ) );
#endif

        if( variant == Box::AR )
            return boundProd( incomingBoxes ).boundSumProd( psi, dest, variant );
        else {
            bool accurate = true;
            Box result;
            if( independent ) {
                size_t nrBoxes = incomingBoxes.size();

                // Prepare multi-index to step through the Cartesian products of 
                // the extreme points corresponding to all disjoint varsets in
                // incomingBoxes
                vector<size_t> ep_dims( nrBoxes, 0 );
                vector<vector<Prob> > ep( nrBoxes );
                for( size_t k = 0; k < nrBoxes; k++ ) {
                    ep[k] = incomingBoxes[k].extremePoints();
                    ep_dims[k] = ep[k].size();
                }

                // For each product of extreme points, calculate its image and
                // update the bound
                Prob image_min( dest.states(), INFINITY );
                Prob image_max( dest.states(), -INFINITY );
                for( MultiFor xi(ep_dims); xi.valid(); xi++ ) {
                    // FIXME: this can be sped up massively by caching the indices
                    Factor prd;
                    for( size_t k = 0; k < nrBoxes; k++ ) {
                        if( variant == Box::ORG )
                            prd *= Factor( incomingBoxes[k].vars(), ep[k][xi[k]] );
                        else if( variant == Box::ALT )
                            prd *= Factor( incomingBoxes[k].vars(), ep[k][xi[k]].normalized() );
                    }
                    prd *= psi;
                    Prob image_ep = prd.marginal(dest,false).p();
                    if( variant == Box::ORG ) {
                        if( image_ep.totalSum() != 0 )
                            image_ep.normalize();
                        else {
                            cerr << "What's going on? incomingBoxes=" << incomingBoxes << ", psi=" << psi << endl;
                            // check whether psi is nice or not...
                            bool nice = true;
                            for( VarSet::const_iterator i = psi.vars().begin(); i != psi.vars().end(); i++ ) {
                                if( psi.marginal(psi.vars() - *i,false).hasZeros() )
                                    nice = false;
                            }
                            cerr << "Hm, psi is " << (nice ? "nice." : "not nice!") << endl;
                            assert( 0 == 1 );
                        }
                    }
                    image_min = min( image_ep, image_min );
                    image_max = max( image_ep, image_max );
                }
                result = Box( dest, image_min, image_max );
            } else { // This can really be a much looser bound...!
                size_t nrIncoming = incomingBoxes.size();
                // delete all simplices in the incoming boxes
                VarSet simplexVars;
                vector<Box> realBoxes;
                for( size_t k = 0; k < nrIncoming; k++ )
                    if( incomingBoxes[k].isSimplex() )
                        simplexVars += incomingBoxes[k].vars();
                    else
                        realBoxes.push_back( incomingBoxes[k] );
                Box prod = boundProd( realBoxes );
                if( prod.vars().states() <= 16 ) {
                    vector<Box> theBoxes;
                    theBoxes.push_back( prod );
                    theBoxes.push_back( Box( simplexVars ) );
                    result = boundSumProd( psi, theBoxes, dest, true, variant );
                    // WAS:  result = prod.boundSumProd( psi, dest, variant );
                } else {
                    cerr << "# Warning: calculating looser bound in the interest of computation time" << endl;
                    accurate = false;
                    result = Box( dest );
                }
            }

#ifdef DEBUG_AR
/*            if( variant == Box::ALT && accurate ) {
                Box resultAR = boundSumProd( psi, incomingBoxes, dest, independent, Box::AR );
                assert( resultAR.contains( result, 1e-9 ) ); // FIXME: this doesn't hold for nonbinary variables
            }
*/
#endif

            return result;
        }
    }


    double Ihler_recursion( double factorStrength, double delta ) {
        double E = (1.0 + factorStrength) / (1.0 - factorStrength);
        if( isinf( delta ) )
            return E;
        else
            return (E * delta + 1.0) / (E + delta);
    }


    double Ihler_distance( const Factor & psi1, const Factor & psi2 ) {
        double dist = 0.0;
        assert( psi1.states() == psi2.states() );
        for( size_t a = 0; a < psi1.states(); a++ )
            for( size_t b = 0; b < psi2.states(); b++ ) {
                double x = (psi1[a] / psi2[a]) / (psi1[b] / psi2[b]);
                if( x > dist )
                    dist = x;
            }
        return sqrt(dist);
    }


    Box Ihler_box( Var v_i, const Factor & psi_i, double delta ) {
        Box box( v_i );
        assert( psi_i.vars() == VarSet(v_i) );
        delta *= delta;
        box.min() = psi_i.p() / (psi_i.p() * (1.0 - delta) + delta);
        box.max() = (psi_i.p() / (psi_i.p() * (delta - 1.0) + 1.0)) * delta;
        return box;
    }

    
    double sgn( double x ) {
        if( x < 0 )
            return -1.0;
        else if( x > 0 )
            return 1.0;
        else
            return 0.0;
    }
    
    double func( double x, double A1, double A2, double B1, double B2, double C1, double C2, double s ) {
        return ((A1 + B1 * x) / (B1 * x + C1) - (A2 + B2 * x) / (B2 * x + C2)) * s;
    }

    // calculate N(psi, i, j)
    // if h is in the box hbox
    // (not optimized yet)
    double FactorStrength( const Factor &psi, const Var &i, const Var &j, const Box &hbox ) {
        assert( psi.vars().contains( i ) );
        assert( psi.vars().contains( j ) );
        assert( i != j );
        assert( hbox.vars() == (psi.vars() - i) );
        VarSet ij(i,j);

        std::vector<Prob> hep = hbox.extremePoints();
//        Factor hextmax;

        double strength = 0.0;
        for( size_t a1 = 0; a1 < i.states(); a1++ )
            for( size_t a2 = 0; a2 < i.states(); a2++ ) 
                if( a1 != a2 ) {
                    Factor psi1 = psi.slice( i, a1 );
                    Factor psi2 = psi.slice( i, a2 );

/*                    for( size_t ep = 0; ep < hep.size(); ep++ ) {
                        Factor hext( hbox.vars(), hep[ep] );
    //                  f = @(h) sum(abs(sum(psi1.*h,2) / sum(sum(psi1.*h)) - sum(psi2.*h,2) / sum(sum(psi2.*h))));
                        cout << ((psi1 * hext).partSum(j) / (psi1 * hext).totalSum() - (psi2 * hext).partSum(j) / (psi2 * hext).totalSum()).abs().totalSum() << endl;
                    }*/

                    double max = 0.0;
            
                    IndexFor b2( j, hbox.vars() );
                    for( size_t b2c2 = 0; b2c2 < hbox.vars().states(); b2c2++, ++b2 ) {
                        for( size_t ep = 0; ep < hep.size(); ep++ ) {
                            Factor hext( hbox.vars(), hep[ep] );
                            hext[b2c2] = 0.0;

                            double B1 = psi1[b2c2];
                            double B2 = psi2[b2c2];
                            double C1 = (psi1 * hext).totalSum();
                            double C2 = (psi2 * hext).totalSum();

//                            denom1 = @(x) (B1*x + C1);
//                            denom2 = @(x) (B2*x + C2);

                            Prob E1 = (psi1 * hext).partSum( j ).p();
                            Prob E2 = (psi2 * hext).partSum( j ).p();
//                          expr = @(x) (E1 / denom1(x) - E2 / denom2(x));
                            Prob crit_x = (E2 * C1 - E1 * C2) / (E1 * B2 - E2 * B1);
                            Prob signs = (E2 * B1 - E1 * B2).sgn();  // sign to the left of crit_x
//                          expr = @(x) ((E1 + B1*x) / denom1(x) - (E2 + B2*x) / denom2(x));   // for b2
                            crit_x[b2] = (E2[b2] * C1 - E1[b2] * C2) / ((E1[b2] - C1) * B2 - (E2[b2] - C2) * B1);
                            signs[b2] = sgn( (C1 - E1[b2]) * B2 - (C2 - E2[b2]) * B1 );

                            std::set<double> set_crit_x;
                            set_crit_x.insert( hbox.min()[b2c2] );
                            for( size_t b1 = 0; b1 < j.states(); b1++ )
                                if( hbox.min()[b2c2] < crit_x[b1] && crit_x[b1] < hbox.max()[b2c2] )
                                    set_crit_x.insert( crit_x[b1] );
                            set_crit_x.insert( hbox.max()[b2c2] );

                            std::set<double>::const_iterator left = set_crit_x.begin();
                            std::set<double>::const_iterator right = left; right++;
                            for( ; right != set_crit_x.end(); left++, right++ ) {
                                if( (*right - *left) > 1e-15 ) {
                                    double center = (*left + *right) / 2.0;
                                    Prob sigs( signs.size(), 0.0 );
                                    for( size_t b1 = 0; b1 < j.states(); b1++ ) {
                                        if( center < crit_x[b1] )
                                            sigs[b1] = signs[b1];
                                        else if( center > crit_x[b1] )
                                            sigs[b1] = -signs[b1];
                                        else
                                            DAI_THROW(INTERNAL_ERROR);
                                    }

                                    double A1 = (E1 * sigs).totalSum() * sigs[b2];
                                    double A2 = (E2 * sigs).totalSum() * sigs[b2];

    //                                numer1 = @(x) (A1 + B1 * x);
    //                                numer2 = @(x) (A2 + B2 * x);
    //                                func = @(x) (sigs(b2) * (numer1(x) / denom1(x) - numer2(x) / denom2(x)));

                                    double f_left = func( *left, A1, A2, B1, B2, C1, C2, sigs[b2] );
                                    double f_right = func( *right, A1, A2, B1, B2, C1, C2, sigs[b2] );
                                    if( f_left > max ) {
                                        max = f_left;
    //                                    hextmax = hext;
    //                                    hextmax[b2c2] = *left;
                                    }
                                    if( f_right > max ) {
                                        max = f_right;
    //                                    hextmax = hext;
    //                                    hextmax[b2c2] = *right;
                                    }

                                    double x1 = (C1*A2-A1*C2 - fabs(C2*B1-C1*B2) * sqrt((C1-A1)*(C2-A2)/(B1*B2))) / (B2*(A1-C1) + B1*(C2-A2));
                                    double f_x1 = func( x1, A1, A2, B1, B2, C1, C2, sigs[b2] );
                                    if( (*left < x1) && (x1 < *right) ) {
                                        if( f_x1 > max ) {
                                            max = f_x1;
    //                                        hextmax = hext;
    //                                        hextmax[b2c2] = x1;
                                        }
                                    }

                                    double x2 = (C1*A2-A1*C2 + fabs(C2*B1-C1*B2) * sqrt((C1-A1)*(C2-A2)/(B1*B2))) / (B2*(A1-C1) + B1*(C2-A2));
                                    double f_x2 = func( x2, A1, A2, B1, B2, C1, C2, sigs[b2] );
                                    if( (*left < x2) && (x2 < *right) ) {
                                        if( f_x2 > max ) {
                                            max = f_x2;
    //                                        hextmax = hext;
    //                                        hextmax[b2c2] = x2;
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if( max > strength )
                        strength = max;
                }


        return 0.5 * strength;
    }


}
