/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __defined_libdai_box_h
#define __defined_libdai_box_h


#include <iostream>
#include <algorithm>
#include <vector>
#include "daialg.h"
#include "weightedgraph.h"
#include "factorgraph.h"


//#define GLPK


namespace libDAI {


    // Box is either a box or a simplex (iff _min is the null vector)
    class Box {
        private:
            VarSet _vars;
        public:
            Prob _min, _max;
            enum Variant {ORG, AR, ALT, IHLER};
        public:
            Box() : _vars(), _min(), _max() {}
            // Construct as simplex
            Box(const VarSet &ns) : _vars(ns), _min(ns.states(), 0.0), _max(ns.states(), 1.0) {}
            Box(const VarSet &ns, double mi, double ma) : _vars(ns), _min(ns.states(), mi), _max(ns.states(), ma) {
                assert(mi <= ma);
            }
            Box(const VarSet &ns, const Prob &mi, const Prob &ma) : _vars(ns), _min(mi), _max(ma) {
                if( !(_min <= _max) )
                    std::cerr << "OOPS: _min = " << _min << ", _max = " << _max << std::endl;
                assert(_min <= _max);
            }
            Box(const VarSet &ns, const std::vector<Prob> &points) : _vars(ns) {
                if( points.size() == 0 ) {
                    double x = 1.0 / ns.states();
                    _min = Prob(ns.states(), x);
                    _max = Prob(ns.states(), x);
                } else {
                    size_t N = ns.states();
                    assert( points[0].size() == N );
                    _min = _max = points[0];
                    for( size_t i = 1; i < points.size(); i++ ) {
                        assert( points[i].size() == N );
                        _min = libDAI::min( _min, points[i] );
                        _max = libDAI::max( _max, points[i] );
                    }
                }
            }
            bool isSimplex() const {
                bool result = true;
                for( size_t i = 0; (i < _min.size()) && result; i++ )
                    if( _min[i] != 0.0 || _max[i] == 0.0 )
                        result = false;
                return result;
            }
            Box & setFull() {
                _min.fill(0.0);
                _max.fill(1.0);
                return *this;
            }
            Box & setEmpty() {
                _min.fill(1.0);
                _max.fill(1.0);
                return *this;
            }
            Prob randomSample() {
                return Prob(_vars.states()).randomize() * (_max - _min) + _min;
            }
            Box( const Box & x ) : _vars(x._vars), _min(x._min), _max(x._max) {}
            Box & operator=( const Box & x ) {
                if( this != &x ) {
                    _vars = x._vars;
                    _min  = x._min;
                    _max  = x._max;
                }
                return *this;
            }

            friend std::ostream& operator<< (std::ostream & os, const Box& c) {
                os << "[" << c._vars << ":" << c._min << " <= " << c._max << "]";
                return( os );
            }
            const VarSet & vars() const { return _vars; }
            const Prob & min() const { return _min; }
            const Prob & max() const { return _max; }
            Prob & min() { return _min; }
            Prob & max() { return _max; }
            bool contains( const Factor & psi, double tol ) const {
                assert( psi.vars() == _vars );
                if( ((_min - tol) <= psi.p()) && (psi.p() <= (_max + tol)) )
                    return true;
                else
                    return false;
            }
            bool contains( const Box & box, double tol ) const {
                assert( box.vars() == _vars );
                if( ((_min - tol) <= box._min) && (box._max <= (_max + tol)) )
                    return true;
                else 
                    return false;
            }
            bool equals( const Box & box, double tol ) const {
                return contains( box, tol ) && box.contains( *this, tol );
            }
            double gap() const {
                if( _max.hasNaNs() || _min.hasNaNs() )
                    return 1.0;
                else
                    return (_max - _min).maxVal();
            }
            Box operator* (const Box &c) const {
//#ifdef LIBDAI_DEBUG
                assert( _min <= _max );
                assert( c._min <= c._max );
//#endif
                assert( _vars == c._vars );
                return Box( _vars, _min * c._min, _max * c._max );
            }
            Box& operator*= (const Box &c) {
//#ifdef LIBDAI_DEBUG
                assert( _min <= _max );
                assert( c._min <= c._max );
//#endif
                assert( _vars == c._vars );
                _min *= c._min;
                _max *= c._max;
                return *this;
            }
            std::vector<Prob> extremePoints() const;
            Box normalize() {
                *this = normalized();
                return *this;
            }
            Box normalized() const;
            Box boundSumProd( const Factor &psi, const VarSet& dest, Box::Variant variant ) const;

            friend Box boundSumProd( const Factor &psi, std::vector<Box> &incomingBoxes, const VarSet &dest, bool independent, Box::Variant variant );

            /// Test for equality (only compares _vars)
            friend bool operator==( const Box &a, const Box &b ) {
                return (a._vars == b._vars);
            }

            /// Test for inequality (only compares _vars)
            friend bool operator!=( const Box &a, const Box &b ) {
                return !(a._vars == b._vars);
            }

            /// Lexicographical comparison (only compares _vars)
            friend bool operator<( const Box &a, const Box &b ) {
                return a._vars < b._vars;
            }
    };

    Box boundProd( const std::vector<Box> &boxes );

    double FactorStrength( const Factor &psi, const Var &i, const Var &j, const Box &hbox );
    double Ihler_recursion( double factorStrength, double delta );
    double Ihler_distance( const Factor & psi1, const Factor & psi2 );
    Box Ihler_box( Var v_i, const Factor & psi_i, double delta );

#ifdef GLPK
    bool TestIfPointInConvexHull( std::vector<Prob> points, Prob point );
#endif


}


#endif
