/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <boost/program_options.hpp>
#include "alldai.h"
#include "bounds.h"
#include "boundprop.h"


using namespace std;
using namespace libDAI;
namespace po = boost::program_options;


void doBound( const string &name, const FactorGraph &fg, size_t limit, vector<double> &gaps, double &tics, size_t maxnodes, bool debug, const JTree &jt, vector<Box> &bounds );

void doBPError( const FactorGraph &fg, size_t limit, vector<double> &errors );
void doCavityBound( const FactorGraph &fg, size_t limit, vector<double> &gaps, vector<Box> &results, bool useSAW, size_t maxnodes, bool debug, double debugtol, const JTree &jt, Box::Variant variant );
void doBoundProp( const FactorGraph &fg, size_t limit, vector<double> &gaps, vector<Box> &results, bool fast, double tol, bool debug, double debugtol, const JTree &jt, size_t verbose );


std::ostream& operator<< (std::ostream& os, const vector<double> &x) {
    os << "[";
    for( size_t i = 0; i < x.size(); i++ )
        os << x[i] << ",";
    os << "]";
    return os;
}
double min( vector<double> & x ) {
    return *min_element( x.begin(), x.end() );
}
double max( vector<double> & x ) {
    return *max_element( x.begin(), x.end() );
}
double avg( vector<double> & x ) {
    double sum = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sum += x[i];
    return sum / x.size();
}
double stddev( vector<double> & x ) {
    double sumsq = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sumsq += x[i] * x[i];
    double _avg = avg( x );
    double result = sumsq / x.size() - _avg * _avg;
    if( result < 0.0 )
        return 0.0;
    else
        return sqrt( result );
}


int main( int argc, char *argv[] ) {
    string filename;
    vector<string> methods;
    bool brief = false;
    size_t limit = 0;
    size_t maxnodes = 1000000000;
    bool debug = false;
    bool dobounds = false;
    bool compare = false;
    size_t offset = 0;

    po::options_description opts_required("Required options");
    opts_required.add_options()
        ("filename", po::value< string >(&filename), "Filename of FactorGraph")
        ("methods", po::value< vector<string> >(&methods)->multitoken(), "Methods to use")
    ;

    po::options_description opts_optional("Allowed options");
    opts_optional.add_options()
        ("help", "Produce help message")
        ("brief", po::value< bool >(&brief), "Give only brief output")
        ("debug", po::value< bool >(&debug), "Check bounds using exact marginals")
        ("bounds", po::value< bool >(&dobounds), "Output bounding boxes")
        ("limit", po::value< size_t >(&limit), "Use only the first n variables")
        ("maxnodes", po::value< size_t >(&maxnodes), "Specify maximum number of nodes used in SAW/Cav")
        ("compare", po::value< bool >(&compare), "Compare methods (outputs one column per method")
        ("offset", po::value< size_t >(&offset), "Offset is added to method index")
    ;

    po::options_description cmdline_options;
    cmdline_options.add(opts_required).add(opts_optional);

    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, cmdline_options), vm);
    po::notify(vm);

    if( vm.count("help") || !(vm.count("filename") && vm.count("methods")) ) {
        cout << "Reads factorgraph <filename> and calculates bounds on its" << endl;
        cout << "(single node) margnals using the methods in <methods>." << endl << endl;
        cout << "The output is brief if <brief> is true; in that case, the ouput consists" << endl;
        cout << "of five columns <methodnr> <methodname> <varlabel> <bound> <time>." << endl;
        cout << "Otherwise, the output consists of (min, max, avg, stddev, time) for each method." << endl;
        cout << "Methods currently implemented are: " << endl;
        cout << "  BP, Cav, CavI, SAW, SAWI, BoundProp" << endl;
        cout << "For Cav* and SAW* one can specify the size of the tree by the --maxnodes option." << endl;
        cout << "If <debug> is true, the bounds are checked with the exact marginals (calculated by JTree)." << endl;
        cout << "If <bounds> is true, the bounds are also reported." << endl;
        cout << "If <compare> is true, compare the various methods." << endl;
        cout << opts_required << opts_optional << endl;
        return 1;
    }

    FactorGraph fg;
    fg.ReadFromFile( filename.c_str() ); 

    if( limit > fg.nrVars() || limit == 0 )
        limit = fg.nrVars();
    if( !vm.count("maxnodes") )
        maxnodes = 1000000000;

//    cerr << "Building Junction Tree..." << endl;
    JTree jt;
    if( debug ) {
        jt = JTree( fg, Properties()("verbose",(size_t)0)("updates",string("HUGIN")) );
        jt.init();
        jt.run();
    }

    vector<vector<double> > results;
    results.reserve( methods.size() );
    for( size_t m = 0; m < methods.size(); m++ ) {
        vector<double> gaps( limit, 1.0 );
        vector<Box> bounds( limit );
        double time = 0;
//        cerr << "Starting " << methods[m] << "..." << endl;
        doBound( methods[m], fg, limit, gaps, time, maxnodes, debug, jt, bounds );
        if( compare ) {
            results.push_back( gaps );
        } else {
            if( brief )
                cout << min(gaps) << " " << max(gaps) << " " << avg(gaps) << " " << stddev(gaps) << " " << time << "     ";
            else {
                for( size_t i = 0; i < limit; i++ ) {
                    cout << m+1+offset << " " << methods[m] << " " << i << " " << gaps[i] << " " << time;
                    if( dobounds )
                        cout << " " << bounds[i];
                    cout << endl;
                }
            }
        }
    }
    if( compare ) {
        cout << "# ";
        for( size_t m = 0; m < methods.size(); m++ ) {
            cout.width( 12 );
            cout << methods[m] << "  ";
        }
        cout << endl;
        cout << "  ";
        for( size_t i = 0; i < limit; i++ ) {
            for( size_t m = 0; m < methods.size(); m++ ) {
                cout.width( 12 );
                cout << results[m][i] << "  ";
            }
            cout << endl;
        }
    }

    return 0;
}



void doBound( const string &name, const FactorGraph &fg, size_t limit, vector<double> &gaps, double &tics, size_t maxnodes, bool debug, const JTree &jt, vector<Box> &bounds ) {
    tics = toc();

    if( name == string("Cav") ) {
        doCavityBound( fg, limit, gaps, bounds, false, maxnodes, debug, 1e-6, jt, Box::ORG );
    } else if( name == string("CavI") ) {
        doCavityBound( fg, limit, gaps, bounds, false, maxnodes, debug, 1e-6, jt, Box::IHLER );
    } else if( name == string("SAW") ) {
        doCavityBound( fg, limit, gaps, bounds, true, maxnodes, debug, 1e-6, jt, Box::ORG );
    } else if( name == string("SAWI") ) {
        doCavityBound( fg, limit, gaps, bounds, true, maxnodes, debug, 1e-6, jt, Box::IHLER );
    } else if( name == string("BoundProp") ) {
        doBoundProp( fg, limit, gaps, bounds, false, 1e-9, debug, 1e-6, jt, 0 );
        // tolerance 1e-9 seems to be too tight in some cases, 1e-8 may be better
        // fast yields loose bounds on Promedas, but identical bounds for pairwise interactions
    }

    tics = toc() - tics;
}


// Calculate actual BP errors
void doBPError( const FactorGraph &fg, size_t limit, vector<double> &errors ) {
    size_t  maxiter = 10000;
    double  tol = 1e-9;
    size_t  verbose = 0;
    Properties opts;
    opts.Set("maxiter",maxiter);
    opts.Set("tol",tol);
    opts.Set("verbose",verbose);

    BP bp(fg, opts("updates",string("SEQMAX")));
    bp.init();
    bp.run();
    
    JTree jt( fg, opts("updates",string("HUGIN")) );
    jt.init();
    jt.run();

    for( size_t i = 0; i < limit; i++ ) {
        Var v_i = fg.var(i);

        double error_i = 0.0;
        for( size_t state = 0; state < v_i.states(); state++ ) {
            double err = fabs( bp.belief(v_i)[state] - jt.belief(v_i)[state] );
            error_i = (err > error_i) ? err : error_i;
        }
        errors[i] = error_i;
    }
}


// New Cavity Bound
void doCavityBound( const FactorGraph &fg, size_t limit, vector<double> &gaps, vector<Box> &results, bool useSAW, size_t maxnodes, bool debug, double debugtol, const JTree &jt, Box::Variant variant ) {
    BP bp;
    vector<Factor> bp_beliefs;

    if( variant == Box::IHLER ) {
        bp = BP(fg, Properties()("maxiter",string("100UL"))("tol",string("1e-9"))("verbose",string("0UL"))("updates",string("SEQFIX")));
        bp.init();
        bp.run();

        bp_beliefs.reserve( fg.nrVars() );
        for( size_t i = 0; i < fg.nrVars(); i++ )
            bp_beliefs.push_back( bp.belief( fg.var(i) ) );
    }

    gaps = vector<double>( limit, 0.0 );
    results = vector<Box>();
    results.reserve( limit );
    for( size_t i = 0; i < limit; i++ ) {
        BoundTree bt;
        bt.buildTree( fg, fg.var(i), maxnodes, useSAW );
//        cerr << "i = " << i << endl;
//        bt.display( cerr, fg );
        Box b = bt.BoxProp( fg, 0, bp_beliefs, variant );
        results.push_back( b );
//        cerr << b << endl;
        gaps[i] = b.gap();
        if( debug ) {
            if( !(b.min().hasNaNs() || b.max().hasNaNs() || jt.belief(fg.var(i)).hasNaNs()) ) {
                if( !b.contains( jt.belief(fg.var(i)), debugtol ) ) {
                    cout << "# ERROR! CavityBound result " << b << " does not contain exact marginal " << jt.belief(fg.var(i)) << endl;
                    cerr << "# ERROR! CavityBound result " << b << " does not contain exact marginal " << jt.belief(fg.var(i)) << endl;
                }
            }
        }
    }
}


// Leisink's BoundPropagation
void doBoundProp( const FactorGraph &fg, size_t limit, vector<double> &gaps, vector<Box> &results, bool fast, double tol, bool debug, double debugtol, const JTree &jt, size_t verbose ) {
    size_t maxcl = 1;
    for( size_t i = 0; i < fg.nrVars(); i++ ) {
        VarSet Di = fg.Delta( fg.var( i ) );
        if( Di.states() > maxcl )
            maxcl = Di.states();
    }

    BoundSet bounds;
    size_t maxiter = 1000;
    BoundProp_solve( fg, fast ? 0 : maxcl, tol, maxiter, bounds, verbose );

    gaps = vector<double>( limit, 1.0 );
    results = vector<Box>();
    results.reserve( limit );
    for( size_t i = 0; i < limit; i++ ) {
        BoundSet::iterator b = find( bounds.begin(), bounds.end(), Box(fg.var(i)) );
        if( b != bounds.end() ) {
            gaps[i] = b->gap();
            results.push_back( *b );
            if( debug ) {
                if( !(b->min().hasNaNs() || b->max().hasNaNs()) ) {
                    if( !b->contains( jt.belief(fg.var(i)), debugtol ) ) {
                        cout << "# ERROR! BoundProp result " << *b << " does not contain exact marginal " << jt.belief(fg.var(i)) << endl;
                    }
                }
            }
        } else {
            cerr << "Not found" << endl;
            return;
        }
    }
}
